/**
  *  \file client/dialogs/searchdlg.cc
  *  \brief Search Dialog
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "client/dialogs/searchdlg.h"
#include "client/dlg-list.h"
#include "client/widgets/searchobj.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/util.h"
#include "ui/widgets/keytranslator.h"
#include "util/nls.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/statictext.h"
#include "game/global.h"
#include "client/widgets/selswitch.h"
#include "game/sel.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/button.h"

/** Global instance of search query. This stores the last search query. */
TUserSearchQuery search_query;

namespace {
    class WSearchDialog : public UIWindow {
        ObjectHolder           h;
        GObjectList&           result;
        WSearchObjectSelector* obj_selector;
        UIStandardListbox*     type_selector;
        UIInputLine*           query_input;
        WObjectSelectionList*  result_list;
        StringList             type_list;
        WScreenId              exit_screen;

        bool prepareQuery(TParsedSearchQuery& q);
        void doMarkResult(bool value);

     public:
        WSearchDialog(GObjectList& result);
        ~WSearchDialog();

        void init();
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);

        void onResultChange();
        void onSearch();
        void onGoTo();

        void doSelectionCommands(UIBaseWidget& btn);

        WScreenId getScreenId() const
            { return exit_screen; }
    };

    enum {
        cm_Search       = 2000,
        cm_GoTo         = 2001,
        cm_Close        = 2002,
        cm_Selection    = 2003,
        cm_Global       = 2004,
        cm_SearchOrGoTo = 2005
    };
}

/** Constructor. Builds the complete search dialog. */
WSearchDialog::WSearchDialog(GObjectList& result)
    : UIWindow(_("Search Object"), BLUE_WINDOW, COLOR_GRAY),
      result(result)
{
    init();
}

/** Destructor. */
WSearchDialog::~WSearchDialog()
{ }

bool
WSearchDialog::prepareQuery(TParsedSearchQuery& query)
{
    search_query.search_kind = TUserSearchQuery::Match(type_selector->getKey());
    search_query.search_obj  = obj_selector->getObjects();
    search_query.query       = query_input->getText();

    return prepareSearchQuery(query, search_query);
}

/** Mark/unmark search result.
    \param value true to mark, false to unmark */
void
WSearchDialog::doMarkResult(bool value)
{
    for (GObjectList::iterator i = result.begin(); i != result.end(); ++i) {
        if (i->isGameObject()) {
            i->getObject().setIsMarked(value);
        }
    }
    getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
}

void
WSearchDialog::init()
{
    /* Initialize types */
    type_list.add(0, _("Name/Id/Comment"));
    type_list.add(1, _("Expression true"));
    type_list.add(2, _("Expression false"));
    type_list.add(3, _("Location"));

    // VBox
    //   HBox 1
    //     VBox 11
    //       Static "Search objects:"
    //       UIFrameGroup 112
    //         WSearchObjectSelector
    //     VBox 12
    //       Static "Search type:"
    //       UIFrameGroup 122
    //         UIStringList
    //       UISpacer
    //   HBox 2
    //     UIInputLine
    //    [UIButton "dropdown"]
    //   HBox 3
    //     UIStaticText(?) status.......
    //     UISpacer
    //     UIButton "Go"
    //   UIFrameGroup [HBox] 4
    //     UIListbox (result)
    //     UIScrollbar
    //   HBox 5
    //     UIButton "Go to"
    //     UIButton "Close"
    //    [UIButton "Selection"]
    //    [UIButton "Global"]
    //     UISpacer
    //    [UIButton "Help"]
    setLayout(UIVBoxLayout::instance5);

    // Group 1
    UIGroup& group_1        = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group_11       = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group_12       = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIFrameGroup& group_112 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    UIFrameGroup& group_122 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));

    obj_selector = &h.add(new WSearchObjectSelector(1000));
    group_112.add(*obj_selector);
    group_11.add(h.add(new UIStaticText(_("Search Objects:"), UIStaticText::StaticTitle)));
    group_11.add(group_112);

    type_selector = &h.add(new UIStandardListbox(1001,
                                                 type_list.getCount(),
                                                 getTextWidth(*fonts[FONT_NORMAL], type_list),
                                                 type_list));
    group_122.add(*type_selector);
    group_12.add(h.add(new UIStaticText(_("Search Type:"), UIStaticText::StaticTitle)));
    group_12.add(group_122);
    group_12.add(h.add(new UISpacer()));

    group_1.add(group_11);
    group_1.add(group_12);
    add(group_1);

    // Group 2
    UIGroup& group_2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    query_input = &h.add(new UIInputLine(INT_MAX, 20, 0, UIInputLine::ilf_Framed, 1002));
    group_2.add(*query_input);
    add(group_2);

    // Group 3
    UIGroup& group_3 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group_3.add(h.add(new UISpacer()));
    group_3.add(h.add(new UIButton(cm_Search, _("Search!"), 0)).addCommand());
    add(group_3);

    // Group 4
    UIFrameGroup& group_4 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    result_list = &h.add(new WObjectSelectionList(1003, 12, 20*font_ems[FONT_NORMAL], result));
    group_4.add(*result_list);
    group_4.add(h.add(new UIScrollbar(*result_list)));
    add(group_4);

    // Group 5
    UIGroup& group_5 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group_5.add(h.add(new UIButton(cm_GoTo,      _("Go to"),     0))           .addCommand());
    group_5.add(h.add(new UIButton(cm_Close,     _("Close"),     SDLK_ESCAPE)) .addCommand());
    group_5.add(h.add(new UIButton(cm_Selection, _("Mark..."),   ss_Alt + 'm')).addCommand());
    // group_5.add(h.add(new UIButton(cm_Global,    _("Global..."),    ss_Alt + 'g')).addCommand());
    group_5.add(h.add(new UISpacer()));
    add(group_5);

    // Administrivia
    add(h.add(new UIFocusIterator(1000, 1003, UIFocusIterator::fi_Tab)));
    add(h.add(new UIKeyTranslator()).addCommand(SDLK_RETURN, cm_SearchOrGoTo));
    add(h.add(new WSelectionSwitcher()));

    pack();

    // Initialize dialog
    obj_selector->setObjects(search_query.search_obj);
    type_selector->scrollToKey(search_query.search_kind);
    query_input->setText(search_query.query);
    query_input->setFocusedWidget();

    onResultChange();
}

bool
WSearchDialog::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch (cmd.code) {
     case cm_GoTo:
        onGoTo();
        return true;
     case cm_Close:
        stop(0);
        return true;
     case cm_Selection:
        doSelectionCommands(sender);
        return true;
     case cm_Global:
        return true;
     case cm_Search:
        onSearch();
        return true;
     case cm_SearchOrGoTo:
        if (result_list->hasState(st_Focused)) {
            onGoTo();
        } else {
            onSearch();
        }
        return true;
     default:
        return UIWindow::handleCommand(cmd, sender);
    }
}

void
WSearchDialog::onResultChange()
{
    result_list->setState(st_Disabled, result.getNumObjects() == 0);
    setStateOf(cm_GoTo,      st_Disabled, result.getNumObjects() == 0);
    setStateOf(cm_Selection, st_Disabled, result.getNumObjects() == 0);
    setStateOf(cm_Global,    st_Disabled, result.getNumObjects() == 0);
}

void
WSearchDialog::onSearch()
{
    // Do we have some objects?
    if (obj_selector->getObjects() == 0) {
        messageBox(_("Please select some object types to search."), _("Search Object"));
        obj_selector->setFocusedWidget();
        return;
    }

    // Prepare query
    TParsedSearchQuery parsed_query;
    search_query.search_kind = TUserSearchQuery::Match(type_selector->getKey());
    search_query.search_obj  = obj_selector->getObjects();
    search_query.query       = query_input->getText();
    if (!prepareSearchQuery(parsed_query, search_query)) {
        return;
    }

    // Clear result list
    result.clear();
    TSearchResultStatus status;

    // Execute query
    executeSearchQuery(parsed_query, status, result);
    getDisplayedTurn().getCurrentUniverse().doScreenUpdates();

    // Update result
    result_list->onListChange(GObjectId());
    onResultChange();

    // Tell user why we didn't find anything
    if (result.getNumObjects() == 0) {
        if (!status.had_object) {
            // No object at all
            messageBox(_("There are no objects of the requested kind."), _("Search Object"));
            obj_selector->setFocusedWidget();
        } else if (status.had_error && !status.did_evaluate && status.last_error.size()) {
            // We didn't evaluate the expression successfully, all invocations failed.
            // This means there's something wrong with the expression.
            messageBox(format(_("Error in search query: %s"), status.last_error), _("Search Object"));
            query_input->setFocusedWidget();
        } else {
            // Nothing found
            messageBox(_("Your query didn't match any object."), _("Search Object"));
            query_input->setFocusedWidget();
        }
    } else {
        result_list->setFocusedWidget();
    }
}

void
WSearchDialog::onGoTo()
{
    // Do we have an object?
    if (!result_list->getCurrentObject().isGameObject())
        return;

    // It is a game object.
    GObject& obj = result_list->getCurrentObject().getObject();
    if (GShip* sh = dynamic_cast<GShip*>(&obj)) {
        // Go to ship screen if played, show it in map otherwise
        if (sh->isPlayed()) {
            exit_screen = WScreenId(WScreenId::ShipScreen, sh->getId());
        } else {
            exit_screen = WScreenId(sh->getPos());
        }
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(&obj)) {
        // If this is a base query, and object is a base, go to base screen.
        // Otherwise, go to planet screen.
        if (pl->isPlayed()) {
            if ((obj_selector->getObjects() & TUserSearchQuery::SearchPlanets) == 0 && pl->hasBase()) {
                exit_screen = WScreenId(WScreenId::BaseScreen, pl->getId());
            } else {
                exit_screen = WScreenId(WScreenId::PlanetScreen, pl->getId());
            }
        } else {
            exit_screen = WScreenId(pl->getPos());
        }
    } else if (GMapObject* mo = dynamic_cast<GMapObject*>(&obj)) {
        // Map object (i.e. anything)
        exit_screen = WScreenId(mo->getPos());
    } else {
        // nothing known
    }
    stop(1);
}

void
WSearchDialog::doSelectionCommands(UIBaseWidget& btn)
{
    enum { Mark, MarkOnly, Unmark };

    StringList list;
    list.add(Mark,     _("Mark found objects"));
    list.add(MarkOnly, _("Mark only found objects"));
    list.add(Unmark,   _("Unmark found objects"));

    UIStandardListbox lb(0, 1, getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
    if (!lb.doStandardMenu(btn))
        return;

    switch (lb.getKey()) {
     case Mark:
        doMarkResult(true);
        break;
     case MarkOnly:
        GSelection().copyTo(getDisplayedTurn().getCurrentUniverse());
        doMarkResult(true);
        break;
     case Unmark:
        doMarkResult(false);
        break;
    }
}

/** Perform search (F7) function. */
WScreenId
doSearch(bool immediate)
{
    GObjectList result;
    WSearchDialog dialog(result);
    ui_root->centerWidget(dialog);
    ui_root->add(dialog);
    if (immediate)
        dialog.onSearch();
    int r = ui_root->run();
    ui_root->close(dialog);
    if (r)
        return dialog.getScreenId();
    else
        return WScreenId();
}

/** Initialize search function. This discards the previous search query. */
void
initSearch()
{
    search_query.search_kind = TUserSearchQuery::MatchName;
    search_query.search_obj  = TUserSearchQuery::AllObjects;
    search_query.query       = "";
}

const ModuleInitializer mod_searchdlg[] = {
    { OnEnterNoGame | OnEnterDirectory | OnEnterGame, initSearch },
    { 0, 0 },
};
