/**
  *  \file client/remote.cc
  *  \brief Remote Control Handling
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "client/remote.h"
#include "game/cmds.h"
#include "game/ship.h"
#include "game/team.h"
#include "util/nls.h"
#include "ui/msgbox.h"

/** Get old remote control state (beginning of turn).
    Returns the logical state of the specified ship. */
TRemoteControlSetting
getOldRemoteControlState(const GShip& sh)
{
    // phost.pas::GetRCFlag
    int rc = sh.getRemoteControlFlag();
    int ow = sh.getOwner();

    /* Special case. Is it used? */
    if (rc == 0)
        rc = ow;

    if (rc < 0) {
        /* Remote control forbidden */
        if (ow == getPlayerId())
            return rc_Forbidden;
        else
            return rc_OtherForbidden;
    } else if (rc == getPlayerId()) {
        /* Remote control flag says it's ours */
        if (ow == getPlayerId())
            return rc_Normal;
        else
            return rc_OurRemoteControlled;
    } else {
        /* Remote control flag says it's someone else's */
        if (ow == getPlayerId())
            return rc_RemoteControlled;
        else
            return rc_Other;
    }
}

/** Get new remote control state (end of turn).
    Returns the logical state after commands have been processed. */
TRemoteControlSetting
getNewRemoteControlState(const GShip& sh, GCommandContainer& cmds)
{
    // phost.pas::GetNewRCFlag
    using std::tolower;
    if (const GCommand* c = cmds.getCommand(GCommand::phc_RemoteControl, sh.getId())) {
        /* Check command */
        const string_t& arg = c->getArg();
        if (arg.size() > 0) {
            switch (tolower(uint8_t(arg[0]))) {
             case 'c': return rc_Applying;
             case 'a': return rc_Normal;
             case 'd': return rc_Dropping;
             default:  return rc_Forbidden;
            }
        } else {
            /* Invalid command */
            return getOldRemoteControlState(sh);
        }
    } else {
        /* No change */
        return getOldRemoteControlState(sh);
    }
}

/** Get indicator color for a remote control setting. */
TFrameColor
getRemoteControlColor(TRemoteControlSetting v)
{
    // phost.par::rc_Colors
    static const TFrameColor colors[] = {
        cf_Red,                 // rc_Forbidden
        cf_None,                // rc_Normal
        cf_Green,               // rc_RemoteControlled
        cf_Green,               // rc_Applying
        cf_Yellow,              // rc_Dropping
        cf_None,                // rc_Other
        cf_Red,                 // rc_OtherForbidden
        cf_None                 // rc_OurRemoteControlled
    };
    return colors[v];
}

/** Remote control. */
void
doRemoteControl(GShip& sh, GCommandContainer& cmds)
{
    /* Remote-control verbs */
    static const char*const rc_verbs[] = {
        "allow",
        "forbid",
        "drop",
        "control",
    };
    /* Mapping of TRemoteControlSetting to verb that was used to enter this state. */
    static const uint8_t rc_verb_index[] = {
        1,                      // rc_Forbidden
        0,                      // rc_Normal
        3,                      // rc_RemoteControlled
        3,                      // rc_Applying
        2,                      // rc_Dropping
        2,                      // rc_Other
        2,                      // rc_OtherForbidden
        0,                      // rc_OurRemoteControlled
    };
    /* Mapping of TRemoteControlSetting to question that can be used to leave this state. */
    static const char*const rc_questions[] = {
        N_("Allow remote control of this ship?"),
        N_("Forbid remote control of this ship?"),
        N_("Drop remote control of this ship?"),
        N_("Cancel request for remote control of this ship?"),
        N_("Cancel dropping remote control of this ship?"),
        N_("Request remote control of this ship?"),
        N_("Request remote control of this ship?"),
        N_("Forbid remote control of this ship?"),
    };

    if (UIMessageBox(_(rc_questions[getNewRemoteControlState(sh, cmds)]), _("Remote Control")).doYesNoDialog()) {
        int new_verb = rc_verb_index[getNewRemoteControlState(sh, cmds)] ^ 1;
        int old_verb = rc_verb_index[getOldRemoteControlState(sh)];

        if (new_verb == old_verb) {
            cmds.removeCommand(GCommand::phc_RemoteControl, sh.getId());
        } else {
            cmds.setCommand(GCommand::phc_RemoteControl, sh.getId(), rc_verbs[new_verb]);
        }

        sh.markDirty();
    }
}
