/**
  *  \file client/widgets/setselect.cc
  *  \brief Class WPlayerSetSelector
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/setselect.h"
#include "cpluslib/format.h"
#include "game/racename.h"
#include "game/struct.h"
#include "util/nls.h"
#include "util/unicodechars.h"

/** Constructor.
    \param id Widget Id
    \param max Maximum number permitted for selection
    \param width Width of the widget in pixels */
WPlayerSetSelector::WPlayerSetSelector(int id, int max, int width)
    : UIListbox(id, max, width, max),
      max(max),
      info()
{ }

WPlayerSetSelector::~WPlayerSetSelector()
{ }

/** Set selected players. */
void
WPlayerSetSelector::setSelectedPlayers(GPlayerSet players)
{
    players &= GPlayerSet::allUpTo(max);
    players -= GPlayerSet(0);
    this->selected_players = players;
    drawWidget();
}

/** Set selectable players. */
void
WPlayerSetSelector::setSelectablePlayers(GPlayerSet players)
{
    players &= GPlayerSet::allUpTo(max);
    players -= GPlayerSet(0);
    this->selectable_players = players;
    drawWidget();
}

/** Set information for a player.
    \param pid Player number
    \param info Information string */
void
WPlayerSetSelector::setInfo(int pid, string_t info)
{
    this->info[pid] = info;
    drawWidget();
}

/** Toggle set of players.
    \param players Players to toggle */
void
WPlayerSetSelector::togglePlayers(GPlayerSet players)
{
    players &= selectable_players;
    if ((selected_players & players) == players)
        setSelectedPlayers(selected_players - players);
    else
        setSelectedPlayers(selected_players + players);
    drawWidget();
}

/** Toggle single player. This also moves the cursor to this player.
    \param player Player number */
void
WPlayerSetSelector::togglePlayer(int player)
{
    if (selectable_players.contains(player)) {
        setSelectedPlayers(selected_players ^ player);
        scrollToPlayer(player);
    }
}

/** Move cursor to player. */
void
WPlayerSetSelector::scrollToPlayer(int player)
{
    scrollTo(player-1, 1);
}

/** Get player number the cursor is on. */
int
WPlayerSetSelector::getCurrentPlayer() const
{
    return getCurrent() + 1;
}

void
WPlayerSetSelector::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    int h    = font_heights[FONT_NORMAL];
    int mid  = 2*font_ems[FONT_NORMAL];
    int mid2 = 16*font_ems[FONT_NORMAL];
    if (mid > extent.w)
        mid = extent.w;
    if (mid2 > extent.w)
        mid2 = extent.w;
    for (int i = from; i < to; ++i) {
        int y = extent.y + i * h;
        int pl = 1 + getTop() + i;
        GfxRect line(extent.x, y, extent.w, h);
        if (!selectable_players.contains(pl)) {
            // Not selectable
            drawBackground(ctx, line);
            ctx.setColor(UIColor::tc_Faded);
        } else if (pl == getCurrent()+1) {
            // Selected
            if (getFocusState() == PrimaryFocus) {
                drawSolidBar(ctx, line, UIColor::tc_Static);
                ctx.setColor(UIColor::tc_InvStatic);
            } else {
                drawBackground(ctx, line);
                ctx.setColor(UIColor::tc_Static);
                drawRectangle(ctx, line);
            }
        } else {
            // Not selected
            drawBackground(ctx, line);
            ctx.setColor(UIColor::tc_Static);
        }

        if (pl > 0 && pl <= max) {
            outTextF(ctx,     line.x,      line.y, mid,         selected_players.contains(pl) ? " " UTF_CHECK_MARK : " " UTF_MIDDLE_DOT);
            if (pl == 12)
                outTextF(ctx, line.x+mid,  line.y, mid2-mid,    _("X - Host"));
            else
                outTextF(ctx, line.x+mid,  line.y, mid2-mid,    format("%X - %s", pl, player_racenames.getShortName(pl)));
            outTextF(ctx,     line.x+mid2, line.y, line.w-mid2, info[pl]);
        }
    }
}

void
WPlayerSetSelector::onItemClick(int item, int x)
{
    /* Click on check mark toggles; click on name or text doesn't */
    if (x < 2*font_ems[FONT_NORMAL])
        togglePlayer(1+item);
}

bool
WPlayerSetSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && hasState(st_Focused)) {
        switch (e.key.code) {
         case ' ':
            togglePlayer(getCurrentPlayer());
            return true;
         case '*':
            togglePlayers(GPlayerSet::allUpTo(NUM_PLAYERS));
            return true;
         case 'a':
            togglePlayer(10);
            return true;
         case 'b':
            togglePlayer(11);
            return true;
         case 'x':
            togglePlayer(12);
            return true;
         default:
            if (e.key.code >= '1' && e.key.code <= '9') {
                togglePlayer(e.key.code - '0');
                return true;
            }
            break;
        }
    }
    return UIListbox::handleEvent(e, second_pass);
}

bool
WPlayerSetSelector::isAccessible(int pos)
{
    return selectable_players.contains(pos+1);
}
