/**
  *  \file client/widgets/specview.cc
  *  \brief Ship Specification View
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/specview.h"
#include "client/widgets/racelist.h"
#include "client/widgets/specbaseinfo.h"
#include "client/widgets/specmaininfo.h"
#include "client/dialogs/funcview.h"
#include "client/widgets/specbuildinfo.h"
#include "game/global.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/keytranslator.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"

/** Constructor. Makes a blank WSpecView that doesn't yet display anything. */
WSpecView::WSpecView()
    : UIGroup(UIVBoxLayout::instance0)
{
    init();
}

/** Destructor. */
WSpecView::~WSpecView()
{ }

/** Create all member widgets. */
void
WSpecView::init()
{
   /* VBox
        UIStaticText [WhiteTitle]
        WSpecBaseInfo
        WSpecMainInfo
        WSpecBuildInfo
        UIStaticText [WhiteNormal]
        HBox
          WRaceList [1,4]
          WRaceList [5,8]
          WRaceList [9,11] */

    w_title = new UIStaticText("", UIStaticText::WhiteTitle);
    add(h.add(w_title));

    w_baseinfo = new WSpecBaseInfo();
    add(h.add(w_baseinfo));

    w_maininfo = new WSpecMainInfo();
    add(h.add(w_maininfo));

    w_buildinfo = new WSpecBuildInfo();
    add(h.add(w_buildinfo));

    add(h.add(new UIStaticText(_("Players:"), UIStaticText::WhiteNormal)));

    UIGroup& hbox = h.add(new UIGroup(UIHBoxLayout::instance5));
    add(hbox);

    static const uint8_t range[] = { 1, 5, 9, 12 };
    for (int i = 0; i < 3; ++i) {
        w_racelist[i] = new WRaceList(100, WRaceList::VerticalLayout | WRaceList::ShowRaceNames | WRaceList::NoColors, range[i], range[i+1]-1);
        hbox.add(h.add(w_racelist[i]));
        w_racelist[i]->sig_click.add(&sig_player_clicked, &Signal1<int>::raise);
    }
}

/** Set ship to display.
    \param query Ship Query describing the ship whose specification you want to see */
void
WSpecView::setQuery(const GShipQuery& query)
{
    /* Change? */
    if (this->query == query)
        return;

    this->query = query;

    /* Update title */
    if (query.getHullId() > 0) {
        w_title->setText(::getHull(query.getHullId()).getName());
    } else {
        w_title->setText(_("No hull selected"));
    }

    /* Update spec infos */
    w_baseinfo->setHull(query.getHullId());
    w_maininfo->setQuery(query);
    w_buildinfo->setHull(query.getHullId());

    /* Update race lists */
    GPlayerSet ps;
    if (query.getHullId() > 0)
        for (int i = 1; i <= NUM_PLAYERS; ++i)
            if (getTruehullSlot(i, query.getHullId()) > 0)
                ps += i;
    for (int i = 0; i < 3; ++i)
        w_racelist[i]->setHighlight(ps);
}

bool
WSpecView::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code == SDLK_F1) {
        showHullFunctions(getDisplayedTurn().getCurrentUniverse(), query);
        return true;
    } else {
        return UIGroup::handleEvent(event, second_pass);
    }
}

/** Display this WSpecView in a standard dialog. */
void
WSpecView::doStandardDialog()
{
    // VBox
    //   WSpecView
    //   HBox
    //     UIButton "OK"
    //     UISpacer
    //     UIButton "F1"
    //     UIButton "W"
    ObjectHolder h;

    UIWindow& w = h.add(new UIWindow(_("Ship Specification"), UIWindow::BLUE_WINDOW, COLOR_DARK));
    w.setLayout(UIVBoxLayout::instance5);

    UIGroup& group = h.add(new UIGroup(UIHBoxLayout::instance5));
    group.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    group.add(h.add(new UISpacer()));
    group.add(h.add(new UIButton("F1", SDLK_F1)));
    w.add(*this);
    w.add(group);
    w.add(h.add(new UIKeyTranslator()).addStop(' ', 1).addStop(SDLK_ESCAPE, 1));
    w.pack();

    ui_root->moveWidgetToEdge(w, 2, 1, 0);
    ui_root->runModal(w);
}
