/**
  *  \file int/vmio.h
  *  \brief Interpreter: Virtual Machine I/O
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_VMIO_H
#define PCC_V2_INT_VMIO_H

#include <map>
#include <functional>
#include "int/bytecode.h"
#include "cpluslib/iterator.h"

class IntExecutionContext;
class IntExecutionFrame;
class IntContext;

/** Context to save a virtual machine state subset.

    Users must call addBCO() and addProcess() to add objects to save. 
    They are actually saved using the save() method. Because there
    cannot be any forward references in a VM file, the addXXX() pass
    gathers objects to save in the correct order, while the save()
    pass just writes out what was gathered so far. This implies that
    there must not be any circular references, which is ensured by the
    compiler.

    Note that this stores a global instance pointer for use by
    serialisation functions; thus is is reentrant, but not
    thread-safe. */
class IntVMSaveContext {
    IntVMSaveContext* prev_instance;

    std::map<IntBCORef,uint32_t> bco_to_id;
    std::vector<IntBCORef> bco_list;
    std::vector<IntExecutionContext*> process_list;

    void     saveBCO(Stream& out, IntBCORef p, uint32_t id);
    void     saveFrame(Stream& out, const IntExecutionFrame& fr);
    void     saveProcess(Stream& out, IntExecutionContext& exc);
 public:
    IntVMSaveContext();
    ~IntVMSaveContext();

    uint32_t addBCO(IntBCORef p);
    void     addProcess(IntExecutionContext& exc);
    void     save(Stream& out);

    static IntVMSaveContext* getCurrentInstance();
};

/** Context to load a virtual machine state subset.

    Note that this stores a global instance pointer for use by
    serialisation functions; thus is is reentrant, but not
    thread-safe. */
class IntVMLoadContext {
    class ObjectLoader;
    IntVMLoadContext* prev_instance;

    std::map<uint32_t,IntBCORef> id_to_bco;
    IntExecutionContext* current_process;

    IntBCORef loadBCO(ObjectLoader&);
    void loadProcess(ObjectLoader&);
    void loadFrames(IntExecutionContext&, Stream& s, uint32_t count);
    void addBCO(IntBCORef p, uint32_t id);
 public:
    IntVMLoadContext();
    ~IntVMLoadContext();

    IntBCORef getBCO(uint32_t id) const;
    IntContext* makeFrameContext(uint32_t id) const;
    void load(Stream& s, bool acceptProcesses);

    static IntVMLoadContext* getCurrentInstance();
};


#endif
