/**
  *  \file ui/window.cc
  *  \brief Top-Level Windows and Root Window
  *
  *  Implementation of the Top-Level Window widget and a root group.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/assert.h"
#include "gfx/clip.h"
#include "gfx/multiclip.h"
#include "gfx/palette.h"
#include "gfx/update.h"
#include "ui/event.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/window.h"
#include "ui/layout.h"
#include "resmgr/resmgr.h"
#include "ui/widgets/cardgroup.h"

/** Draw frames around current window's widgets. This is used to
    visualize layout management.
    \param can Screen canvas
    \param p   Root widget

    Widgets are color-coded as follows:
    - red: UIGroup (all descendants shown)
    - white: UICardGroup (only visible descendant shown)
    - green, slashed: UISpacer
    - yellow: regular widgets */
static void
doFrames(GfxCanvas& can, UIBaseWidget& p)
{
    if (!p.hasState(UIWidget::st_Exposed))
        return;

    GfxContext ctx;
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useCanvas(can);
    if (UICardGroup* c = dynamic_cast<UICardGroup*>(&p)) {
        if (UIBaseWidget* ch = c->getFocusedChild())
            doFrames(can, *ch);
        ctx.setColor(COLOR_WHITE);
    } else if (UIBaseWidget* pp = p.getFirstChild()) {
        while (pp != 0) {
            doFrames(can, *pp);
            pp = pp->getNextSibling();
        }
        ctx.setColor(COLOR_RED);
    } else if (dynamic_cast<UISpacer*>(&p)) {
        ctx.setFillPattern(LTSLASH_FILL);
        ctx.setColor(COLOR_GREEN);
        drawBar(ctx, p.getExtent());
    } else {
        ctx.setColor(COLOR_YELLOW);
    }
    drawRectangle(ctx, p.getExtent());
}

static bool
doLayoutQuery(GfxCanvas& can, UIBaseWidget& p, GfxPoint pt)
{
    /* Only exposed widgets, if mouse is inside widget, and it's not a spacer */
    if (!p.hasState(UIWidget::st_Exposed)
        || !p.isInside(pt)
        || dynamic_cast<UISpacer*>(&p) != 0)
    {
        return false;
    }

    /* Set up graphics */
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    /* Check whether user clicked one of our children */
    bool hadChildren = false;
    if (UICardGroup* c = dynamic_cast<UICardGroup*>(&p)) {
        if (UIBaseWidget* ch = c->getFocusedChild())
            hadChildren = doLayoutQuery(can, *ch, pt);
        ctx.setColor(COLOR_WHITE);
    } else if (UIBaseWidget* pp = p.getFirstChild()) {
        while (pp != 0) {
            hadChildren |= doLayoutQuery(can, *pp, pt);
            pp = pp->getNextSibling();
        }
        ctx.setColor(COLOR_RED);
    } else {
        ctx.setColor(COLOR_YELLOW);
    }

    if (!hadChildren) {
        UIBaseWidget::LayoutInfo li;
        p.getLayoutInfo(li);

        // minimum size
        ctx.getCanvas().drawBar(GfxRect(p.getExtent().getTopLeft(), li.min_size),
                                ctx.getRawColor(), GFX_TRANSPARENT, GRAY50_FILL, 64);

        // preferred size
        ctx.setLinePattern(0x77);
        drawRectangle(ctx, GfxRect(p.getExtent().getTopLeft(), li.pref_size));

        // actual size
        ctx.setLinePattern(0x88);
        drawRectangle(ctx, p.getExtent());
    }
    return true;
}

static bool
simplifyEvent(UIEvent& event)
{
    if (event.type == ev_Key && (event.key.code & ss_Shift) != 0) {
        event.key.code &= ~ss_Shift;
        return true;
    } else if (event.type == ev_Key && (event.key.code >= 'A' && event.key.code <= 'Z')) {
        event.key.code += 'a'-'A';
        return true;
    } else {
        return false;
    }
}

/**
  *  \class UIRoot
  *  \brief Root Group
  *
  *  This implements a ``root group'': a group which does not need
  *  a owner group.
  *
  *  This particular incarantion supports:
  *  - overlapping windows
  *  - painting on a GfxCanvas (which it expects to own completely)
  *  - nested invocations of the event loop
  */

/** Create Root Group on Canvas. The canvas should be a GfxSurface
    or something similar. */
UIRoot::UIRoot(Ptr<GfxCanvas> screen)
    : UIBaseWidget(),
      multi_clip(*screen),
      screen(screen)
{
    setSkin(UISkin::getDefaultSkin());
    setExtent(GfxRect(*screen), true);
    setState(st_Exposed | st_Focused | st_Selected, true);
}

UIRoot::~UIRoot()
{ }

/** Perform all pending redraws (see drawWidget()). */
void
UIRoot::performDeferredRedraws()
{
    performRegionUpdates();

    if (hasState(st_WantRedraw | st_ChildRedraw)) {
        // FIXME: lock/unlock?
        GfxUpdateOptimizer updater(*screen);
        updater.lock();
        performDeferredDraw(updater);
        updater.unlock();
        updater.doUpdates();
    }
}

void
UIRoot::performRegionUpdates()
{
    if (!multi_clip.empty()) {
        // FIXME: lock/unlock?
        GfxUpdateOptimizer updater(multi_clip);
        updater.lock();
        drawEverything(updater);
        updater.unlock();
        updater.doUpdates();
        multi_clip.clear();
    }
}

void
UIRoot::drawContent(GfxCanvas& canvas)
{
    // formerly drawBackground()
    canvas.drawBar(getExtent(), standard_colors[COLOR_BLACK], GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);
    canvas.update(getExtent());
}

void
UIRoot::drawEverything(GfxCanvas& canvas)
{
    GfxMultiClip mc(canvas);
    mc.add(GfxRect(*screen));

    for (UIBaseWidget* p = getFirstChild(); p; p = p->getNextSibling()) {
        p->drawEverything(mc);
        mc.remove(p->getExtent());
    }
    if (mc.isVisible(getExtent())) {
        drawContent(mc);
    }
}

void
UIRoot::drawBelow(UIBaseWidget& child, GfxCanvas& canvas)
{
    ASSERT(child.getParent() == this);

    performRegionUpdates();

    GfxMultiClip mc(canvas);
    mc.add(GfxRect(*screen));
    for (UIBaseWidget* p = child.getNextSibling(); p; p = p->getNextSibling()) {
        p->drawEverything(mc);
        mc.remove(p->getExtent());
    }
    if (mc.isVisible(getExtent())) {
        drawContent(mc);
    }
}

GfxCanvas&
UIRoot::getCanvasFor(UIBaseWidget& child, ObjectHolder& h)
{
    ASSERT(child.getParent() == this);

    performRegionUpdates();

    if (&child == getFirstChild()) {
        return *screen;
    } else {
        GfxMultiClip& mc = h.add(new GfxMultiClip(*screen));
        mc.add(GfxRect(*screen));
        for (UIBaseWidget* p = getFirstChild(); p != 0 && p != &child; p = p->getNextSibling()) {
            mc.remove(p->getExtent());
        }
        return mc;
    }
}

/** Event Handling. This does the focus policy:
    - keyboard and command events go to topmost window only;
    - mouse events go to all windows that can see it;
    - rest (timer, broadcast) goes to everyone. */
bool
UIRoot::handleEvent(const UIEvent& e, bool second_pass)
{
    static bool pending_layout_query = false;
    static bool suppress_mouse_events = false;

    /* events we do ourselves */
    if (e.type == ev_Key) {
        if (e.key.code == ss_Ctrl + ss_Shift + 's') {
            doScreenshot();
            return true;
        }
        if (e.key.code == ss_Ctrl + ss_Shift + 'q') {
            putKey(vk_Quit);
            return true;
        }
        if (e.key.code == ss_Ctrl + ss_Shift + 'f') {
            if (UIBaseWidget* p = getFirstChild()) {
                doFrames(*screen, *p);
                screen->update(getExtent());
            }
            return true;
        }
        if (e.key.code == ss_Ctrl + ss_Shift + 'l') {
            pending_layout_query = !pending_layout_query;
            return true;
        }
        if (e.key.code == ss_Ctrl + ss_Shift + 'r') {
            drawWidget(true);
            return true;
        }
    }

    /* propagate stuff */
    if (e.type == ev_Key) {
        /* top widget only if modal, propagate otherwise. */
        for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling()) {
            if (p->handleEvent(e, second_pass))
                return true;
            if (p->hasState(st_Modal))
                return false;
        }
        return false;
    }
    if (e.type == ev_Mouse) {
        int mx = e.mouse.x;
        int my = e.mouse.y;
        /* Do we want to handle it ourselves? */
        if (pending_layout_query && e.mouse.b != 0) {
            if (UIBaseWidget* p = getFirstChild()) {
                doLayoutQuery(*screen, *p, GfxPoint(mx, my));
                screen->update(getExtent());
            }
            pending_layout_query = false;
            suppress_mouse_events = true;
            return true;
        }
        /* Suppress mouse events means to wait for button release */
        if (suppress_mouse_events) {
            if (e.mouse.b == 0) {
                suppress_mouse_events = false;
            } else {
                return true;
            }
        }
        /* Regular event distribution, to widgets which see it only */
        for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling()) {
            if (p->handleEvent(e, second_pass))
                return true;
            if (p->isInside(GfxPoint(mx, my)) || p->hasState(st_Modal))
                return false;
        }
        return false;
    }
    /* other widgets: broadcast */
    for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling())
        if (p->handleEvent(e, second_pass))
            return true;

    return false;
}

void
UIRoot::onAddRemoveChild(UIBaseWidget& the_widget, bool added)
{
    if (added) {
        // FIXME?
        multi_clip.remove(the_widget.getExtent());
        performRegionUpdates();
        setFocusedWidget(the_widget);
        setSelectedWidget(the_widget);
    } else {
        // FIXME?
        multi_clip.add(the_widget.getExtent());
    }
}

void
UIRoot::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_NoLayout;
}

void
UIRoot::onResize(GfxRect /*oldsize*/, bool /*by_layout*/)
{ }

/** Handle Child Resize.
    \todo Right now, this is almost a copy of UIGroup::handleChildResize(). */
void
UIRoot::onChildResize(UIBaseWidget& child, GfxRect oldsize)
{
    performRegionUpdates();
    if (child.getExtent().contains(oldsize)) {
        /* child completely covers old position. */
        child.drawWidget();
    } else {
        /* child has moved. Complete redraw in case something previously
           obscured (i.e. background) now becomes visible. */
        GfxMultiClip mc(*screen);
        mc.add(child.getExtent());
        mc.add(oldsize);

        // FIXME: lock/unlock?
        GfxUpdateOptimizer updater(mc);
        drawEverything(updater);
        updater.doUpdates();
    }
}

int
UIRoot::run()
{
    while (pending_stops.size() == 0) {
        UIEvent e;
        if (!haveQueuedEvent())
            performDeferredRedraws();
        getEvent(e, hasState(st_IdleNotify));

        /* Process event. If nobody wants it, simplify it. */
        bool ok = false;
        do {
            if (handleEvent(e, false) || handleEvent(e, true))
                ok = true;
        } while (!ok && simplifyEvent(e));

        /* If it is an unhandled user event, clear the prefix
           argument. User events are keypresses and mouse button
           releases. Since prefix argument entry is terminated
           by either a keypress, or a pressed mouse button, it
           suffices to test for b==0. This either means a button
           release as required, or we don't have an argument. */
        if (!ok)
            if (e.type == ev_Key || (e.type == ev_Mouse && e.mouse.b == 0))
                consumePrefixArg();
    }
    int result = pending_stops.front();
    pending_stops.pop();
    return result;
}

void
UIRoot::stop(int arg)
{
    pending_stops.push(arg);
}

// New functions:

void
UIRoot::add(UIBaseWidget& w)
{
    addChildWidget(w, 0);
    setFocusedWidget(w);
}

void
UIRoot::close(UIBaseWidget& w)
{
    removeChildWidget(w);
}

/** Run widget. The widget usually is a modal window (but needn't be).
    This function call combines the add(), run() and close() calls in
    one call. */
int
UIRoot::runModal(UIBaseWidget& w)
{
    add(w);
    int result = run();
    close(w);
    return result;
}

/** Center Widget on Screen. */
void
UIRoot::centerWidget(UIBaseWidget& w)
{
    // FIXME: could be member of UIBaseWidget
    GfxRect ext = w.getExtent();
    ext.centerWithin(getExtent());
    w.setExtent(ext);
}

/** Move widget to a "logical" position. The position is given by two
    alignment values, as for text alignment.
    \param w      corpus delicti
    \param x_pos  horizontal position: 0=left, 1=center, 2=right
    \param y_pos  vertical position: 0=top, 1=center, 2=right
    \param offset distance to edge. At offset 10, the window will stay
                  10 pixels from screen border. */
void
UIRoot::moveWidgetToEdge(UIBaseWidget& w, int x_pos, int y_pos, int offset)
{
    // FIXME: could be member of UIBaseWidget
    GfxRect ext = w.getExtent();
    ext.moveToEdge(getExtent(), x_pos, y_pos, offset);
    w.setExtent(ext);
}

/** Schedule region for redraw. Use to perform redraw after having messed
    with the screen outside the regular UI kernel. */
void
UIRoot::markRedrawRegion(GfxRect r)
{
    multi_clip.add(r);
}

/** Process Timer Event. Post timer event to topmost window, if needed.
    Call this occasionally while doing expensive computations. This
    permits display of `idle' indicators.

    \note that this is different from standard timer events which are
    posted to all windows. This one goes to the topmost window only
    for performance reasons. Wo don't want to degrade the performance
    of the above-mentioned heavy processing by stupid event handling. */
void
UIRoot::postTimerEvent()
{
    static long old_timer = 0;
    long this_timer = getTimer();

    if (this_timer == old_timer)
        return;

    old_timer = this_timer;

    UIEvent e;
    e.type = ev_Tick;
    if (UIBaseWidget* first = getFirstChild()) {
        first->handleEvent(e, false) || first->handleEvent(e, true);
        performDeferredRedraws();
    }
}

/****************************** UIWindowSkin *****************************/

UIWindowSkin::UIWindowSkin(int bs)
    : UISkin(bs == COLOR_GRAY ? GRAY_COLOR_SET :
             bs == COLOR_DARK ? DARK_COLOR_SET :
             bs == COLOR_BLACK ? BLACK_COLOR_SET : GRAY_COLOR_SET)
{ }

/******************************** UIWindow *******************************/

/**
  *  \class UIWindow
  *  \brief Standard Window
  *
  *  This is a standard framed window.
  */

Ptr<GfxPixmap> UIWindow::frame_blue;

UIWindow::UIWindow(string_t name, int fs, int bs, int border)
    : UILayoutableWidget(UINullLayout::instance),
      skin(bs),
      name(name),
      frame_style(fs),
      bg_style(bs),
      border_space(border)
{
    if (!frame_blue.ptr())
        frame_blue = requestPixmap("winblue");
    setSkin(skin);
    setState(st_Modal, true);
}

void
UIWindow::drawContent(GfxCanvas& canvas)
{
    drawWindow(canvas, getExtent(), bg_style, name, true);
    canvas.update(getExtent());
}

void
UIWindow::adjustSize(GfxRect& size, bool to_inner)
{
    if (to_inner) {
        // window dimensions to client size
        size.w -= 8 + 2*border_space;
        size.h -= 28 + 2*border_space;
        size.x += 4 + border_space;
        size.y += 24 + border_space;
    } else {
        // client size to window dimensions
        size.w += 8 + 2*border_space;
        size.h += 28 + 2*border_space;
        size.x -= 4 + border_space;
        size.y -= 24 + border_space;

        // ensure minimum size so that window title fits
        int min_w = fonts[FONT_TITLE]->getTextWidth(name) + 6;
        if (size.w < min_w)
            size.w = min_w;
    }
}

/** Draw a window. Available as static function to be callable from
    outside (widgets that look like windows but aren't, like the VCR
    screen) */
void
UIWindow::drawWindow(GfxCanvas& canvas, const GfxRect& extent,
                     int bg_style, string_t name, bool bg)
{
    drawFrameUp(canvas, extent);
    drawTiledArea(canvas, GfxRect(extent.x + 1, extent.y + 1,
                                  extent.w - 2, 22),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    drawTiledArea(canvas, GfxRect(extent.x + 1, extent.y + 23,
                                  2, extent.h - 26),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    drawTiledArea(canvas, GfxRect(extent.x + extent.w - 3, extent.y + 23,
                                  2, extent.h - 26),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    drawTiledArea(canvas, GfxRect(extent.x + 1, extent.y + extent.h - 3,
                                  extent.w - 2, 2),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);

    drawFrameDown(canvas, GfxRect(extent.x + 3, extent.y + 23,
                                  extent.w - 6, extent.h - 26));
    if (bg) {
        // synchronize this with UIWindowSkin::drawBackground
        canvas.drawBar(GfxRect(extent.x + 4, extent.y + 24,
                               extent.w - 8, extent.h - 28),
                       standard_colors[bg_style],
                       GFX_TRANSPARENT,
                       SOLID_FILL,
                       GFX_ALPHA_OPAQUE);
    }

    GfxContext ctx;
    ctx.useCanvas(canvas);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_WHITE);
    ctx.setTextAlign(1, 0);
    ctx.useFont(*fonts[FONT_TITLE]);
    outTextF(ctx, extent.x + extent.w/2, extent.y + 2, extent.w, name);
}
